//++++++++++++++++++++++++++++++++++
// +++++ DDEAdvise for MATLAB 16bit

#include <windows.h>
#include <ddeml.h>
#include <string.h>		//fuer memmove()
#include <stdlib.h>

#define DLLMEX
extern "C"
{
	#include "..\header\mex.h"
	#include "..\header\testargs.h"
}

#include "..\header\fastdde.h"
#include "..\header\i_field.h"	//Klasse fuer Advise-Tabelle

#pragma hdrstop



#define ARGS_IN_MIN 3
#define ARGS_IN_MAX 6

#define ARG1_IN		prhs[0]		//channel
#define ARG2_IN		prhs[1]		//item
#define ARG3_IN		prhs[2]		//CallBack-Function
#define ARG4_IN		prhs[3]		//upmtx UpdateMatrix
#define ARG5_IN		prhs[4]		//cbFormat
#define ARG6_IN		prhs[5]		//timeout



#define ARG1_OUT	plhs[0]		//return code



//globales (Advise)-Tabellen-Objekt 
//speichert alle Advise-Link-spezifischen Informationen (hConv, hsz, Callbackstring, Matrixname)
InfoField AdvTable;



extern "C" void mexFunction(
		  int     nlhs,
		  Matrix  *plhs[],
		  int     nrhs,
		  Matrix  *prhs[]
		  )

{
	DWORD	timeout			= TIMEOUT;
	WORD	cbFormat		= CF_TEXT;	//default
	BOOL	StrFormatChk    = TRUE;		//zweiten Wert im Format-Argument testen

    double	Result		= 0;

	char*	szDestMatrix = NULL;


	if ( nrhs >= ARGS_IN_MAX - 2 )							//4. Parameter (upmtx) angegeben?
		szDestMatrix = TestUpmtxArg( ARG4_IN );
	if ( nrhs >= ARGS_IN_MAX - 1 )							//5. Parameter (Clipboard-Format)?
		cbFormat = TestFormatArg( ARG5_IN, &StrFormatChk );
	else
		StrFormatChk = FALSE;
				
	if ( nrhs == ARGS_IN_MAX )								//6. Parameter (TimeOut)?		
		timeout = TestTimeoutArg( ARG6_IN );


	if (	TestIOArgs( nrhs, ARGS_IN_MIN, ARGS_IN_MAX ) &&		//alle erforderlichen Parameter angegeben?
			TestChannelArg( ARG1_IN ) &&						//Channel Argument ok?
			TestItemArg( ARG2_IN ) &&
			TestCallbackArg( ARG3_IN ) &&
			szDestMatrix != (char*)-1 &&
			cbFormat != (WORD)-1 &&								//sind bei ungueltigen Argumenten
			timeout != (DWORD)-1	)							// -1
	{

		DWORD	hChannel[2];	//2 mal 32 bit
		int		n;


		char* szItem = (char*) mxCalloc( n= (int)mxGetN(ARG2_IN) + 1, sizeof(char) );
		mxGetString( ARG2_IN, szItem, n );

		char* szCallback = (char*) mxCalloc( n= (int)mxGetN(ARG3_IN) + 1, sizeof(char) );
		mxGetString( ARG3_IN, szCallback, n );

        hmemcpy( (char*)hChannel, (char huge*)mxGetPr(ARG1_IN), sizeof(double) );

		DWORD	idInst	= hChannel[0];
        HCONV	hConv	= hChannel[1];
		HSZ		hszItem	= DdeCreateStringHandle( idInst, szItem, CP_WINANSI );
				
        if ( hszItem )
        {
		    HDDEDATA	hTrans;

			AdvTable.AddEntry( hConv, hszItem, szCallback, szDestMatrix, StrFormatChk );	//Advise-Link-Tabelle aktualisieren

			if ( szDestMatrix )	//HotLink
				hTrans = DdeClientTransaction( (LPBYTE)NULL, (DWORD)0, hConv, hszItem, cbFormat,
													(XTYP_ADVSTART /*| XTYPF_ACKREQ*/), timeout, NULL );
			else				//WarmLink
				hTrans = DdeClientTransaction( (LPBYTE)NULL, (DWORD)0, hConv, hszItem, cbFormat,
													(XTYP_ADVSTART | XTYPF_NODATA /*| XTYPF_ACKREQ*/), timeout, NULL );
		//hTrans = 1, wenn o.k.
			if ( hTrans )
				Result = (double)1;


			DdeFreeStringHandle ( idInst, hszItem );
		}

		mxFree( szCallback );
		mxFree( szItem );
		
		if ( szDestMatrix )
			mxFree( szDestMatrix );

		ARG1_OUT = mxCreateFull(1, 1, REAL);
		hmemcpy( (char huge*)mxGetPr(ARG1_OUT), (char*)&Result, sizeof(double) );
	}
}







/////////////////////////////////////////////////////////////////////////////////////////////////////
extern "C" void FAR PASCAL __export DDEAdvise( HCONV hConv, HSZ hszItem, HDDEDATA hData, WORD cbFormat )
{

	if ( hData )
	{
		//MATLAB-Matrix erzeugen	
		if ( cbFormat == CF_TEXT )
		{
			if ( AdvTable.IsStringMatrix(hConv, hszItem) )
				AdvCFTextStringReq( hData, AdvTable.GetDestMatrix(hConv, hszItem) );
            else
            {
				if ( !AdvCFTextReq( hData, AdvTable.GetDestMatrix(hConv, hszItem) ) )
					mexPrintf("Memory allocation error. Matrix not actualized.\r\n");
			}
		}
		else	//Excel Format
		{
			if ( !AdvFastDDEReq( hData, AdvTable.GetDestMatrix(hConv, hszItem) ) )
				mexPrintf("Error in Data. Matrix not actualized.\r\n");
		}

	}

	
//MATLAB-Befehl ausfuehren (Warm- und HotLink)
	char* szCallbackStr;


	if ( hszItem == (HSZ)NULL )		//falls hszItem = NULL, werden alle AdviseLinks, 
	{								//die hConv entsprechen, bearbeitet.
		int index = 0;
		
		do
		{
			szCallbackStr = AdvTable.exGetCallback( hConv, &index );
			if ( szCallbackStr )
				mexEvalString( szCallbackStr );
		}
		while ( szCallbackStr != NULL );
	}

	else
	{
		szCallbackStr = AdvTable.GetCallback( hConv, hszItem );
		if ( szCallbackStr )
			mexEvalString( szCallbackStr );
	}

}



/////////////////////////////////////////////////////////////////////////////////////////////////////
extern "C" BOOL FAR PASCAL __export DDEUnadvise( HCONV hConv, HSZ hszItem )
{

	return AdvTable.DelEntry( hConv, hszItem );

}




/////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL AdvFastDDEReq( HDDEDATA hData, char* szDestMatrix ) 
{
	TableHeader	TabHeader;
	

	BYTE* hDataBlock = DdeAccessData( hData, NULL );
	
//tdtTable-Header lesen
    hmemcpy( (char*)&TabHeader, (char*)hDataBlock, sizeof(TabHeader) );
    hDataBlock += sizeof(TabHeader);
    

//tdtFloat-Header entfernen

	if ( !RemoveHeader( &TabHeader, hDataBlock ) )
	{
		DdeUnaccessData( hData );
    	return FALSE;
    }

//Matrix transponieren
	double value;
    DWORD index;

	double* src		= (double*)hDataBlock;
	double huge* dest	= (double huge*)mxCalloc( (WORD)TabHeader.rows * (WORD)TabHeader.columns, sizeof(double) ); 
	double huge* dest_bak	= dest;	//Zeiger auf den Anfang retten
	
	for ( int i = 0; i < (int)TabHeader.columns; i++ )
	{
		for ( int k = 0; k < (int)TabHeader.rows; k++ )
		{
            index = (DWORD)i+ ((DWORD)k * (DWORD)TabHeader.columns);
			value = src[index];
			hmemcpy( (char huge*)dest, (char*)&value, sizeof(double) );
			dest++;
		}
	}

//Matrix erzeugen
	mexPutFull( szDestMatrix, (int)TabHeader.rows, (int)TabHeader.columns, dest_bak, NULL );    
    mxFree( dest_bak );
	
	DdeUnaccessData( hData );


	return TRUE;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL RemoveHeader( TableHeader* TabHeader, BYTE huge* src )
{
//Groesse der Matrix in bytes
	DWORD		Size		= (DWORD)TabHeader->rows * (DWORD)TabHeader->columns * sizeof(double);
	DWORD		BytesMoved	= (DWORD)0;
	BOOL		result		= TRUE;
	BYTE huge*	dest		= src;

	FloatHeader	FtHeader;

	while ( BytesMoved < Size )
    {
	    hmemcpy( (char*)&FtHeader, (char huge*)src, sizeof(FtHeader) );
        src += sizeof(FtHeader);
		if ( FtHeader.Float != tdtFloat )
		{
			result = FALSE;
			break;
		}

        
		memmove( (char*)dest, (char*)src, FtHeader.cbFloat);

        src += FtHeader.cbFloat;
        dest += FtHeader.cbFloat;
        BytesMoved += FtHeader.cbFloat;
    }
    
    return result;
}





/////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL AdvCFTextReq( HDDEDATA hData, char* szDestMatrix )
{
	HGLOBAL	hGlobal;
	int		rows, columns;

	char* hDataString = (char*)DdeAccessData( hData, NULL );

	GetMatrixSize( hDataString, &rows, &columns );
	DWORD matrixSize = (DWORD)rows * (DWORD)columns;

//Speicher reservieren
	hGlobal = GlobalAlloc( GPTR, matrixSize * sizeof(double) );
    double huge* destTransp_bak	= (double huge*)mxCalloc( (WORD)rows * (WORD)columns, sizeof(double) );
    											//mxCalloc ( wegen mexPutFull() )
	if ( !hGlobal || !destTransp_bak )
	{
		DdeUnaccessData( hData );
		DdeFreeDataHandle( hData );
		return FALSE;
	}

    double* src	= (double*) GlobalLock( hGlobal );	//in src steht nach Aufruf von 

	if ( Str2Double( src, hDataString ) )			//Str2Double() die untransponierte Matrix
	{
	    DWORD index;
		double value;
		double huge* dest	= destTransp_bak; 
	
//Matrix transponieren
		for ( int i = 0; i < columns; i++ )
		{
			for ( int k = 0; k < rows; k++ )
			{
	            index = (DWORD)i+ ((DWORD)k * (DWORD)columns);
				value = src[index];
				hmemcpy( (char huge*)dest, (char*)&value, sizeof(double) );
				dest++;
			}
		}

//Matrix erzeugen
		mexPutFull( szDestMatrix, rows, columns, destTransp_bak, NULL );    
	}
	mxFree( destTransp_bak );

	GlobalUnlock( hGlobal );
	GlobalFree( hGlobal );

	DdeUnaccessData( hData );
	DdeFreeDataHandle( hData );

	return TRUE;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
void GetMatrixSize( char huge* hDataString, int* rows, int* columns )
{
	DWORD i = 0;

	*rows = 0;
	*columns = 1;

	while ( hDataString[i++] )
	{
		if ( hDataString[i] == '\t' && *rows == 0 )
			(*columns)++;
	    if ( hDataString[i] == '\r' )
		    (*rows)++;
	}
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL Str2Double( double huge* dest, char huge* hDataString )
{
	DWORD i = 0;
	
	while ( *hDataString != '\0' )
	{
		if ( *hDataString == '\t' || *hDataString == '\r' )
		{
			mexPrintf( "Matrix not actualized. Possible empty element of source matrix.\r\n" );
			return FALSE;
		}
		else	
			dest[i] = atof( hDataString );
		
		Go2Next( &hDataString );
		i++;	
	}

    return TRUE;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
_inline void Go2Next( char huge* huge* hDataString )
{
	do
	{
		switch ( **hDataString )
		{
			case '\t':	(*hDataString)++;
						return;
			case '\r':	(*hDataString) += 2;
						return;
		}
	}
	while ( (*hDataString)++ );
}




BOOL AdvCFTextStringReq( HDDEDATA hData, char* szDestMatrix )
{
	char* hDataString = (char*)DdeAccessData( hData, NULL );
	Matrix* pm;

	pm = mxCreateString( hDataString );
	mxSetName( pm, szDestMatrix );
	mexPutMatrix( pm );

	return TRUE;
}